#!/usr/bin/php -qCdregister_argc_argv=1
<?php
/**
 * api: phrep
 * type: main
 * title: PHreProcessor
 * description: Rewrites preprocessing statements in PHP source files
 * url: http://fossil.include-once.org/phrep/
 * version: 0.2.9
 * license: MITL
 * author: mario@include-once.org
 * category: devel
 * depends: php (>= 5.4)
 * suggests: bin:make
 * config: { type: env list colon, name: PHP_INCLUDE_PATH, description: Additional include search dirs. }
 * architecture: all
 * pack: phrep=/usr/bin/, *.1=/usr/share/man/man1/
 * 
 * This tool works similar to CPP, in that it rewrites preprocessing directives
 * like #define and #include in source files. It's designed for PHP scripts
 * however, is context-sensitive per default, and does not strip other comments.
 * 
 * Usage:
 *     phrep  -D PKG_VERSION=1.2.3  -i infile.src -o outfile.php
 *
 * While #directives are honored regardless of file location, constant and macro
 * replacements are only applied within <?php code sections, in the -m token mode.
 *
 */


namespace {

  #-- Init
#ifndef PKG_COMBINED
  include(__DIR__."/macro.php");
#endif
  $proc = new io\MacroProcessor();
  $opts = new ShellFlags(getopt("i:o:I:D:m:p:wWhvVkfEDn:", ShellFlags::$map), $proc);
  empty($opts->config) and $opts->config();

  #-- Process main file
  if (empty($opts->input)) {
    $proc->stderr(1, "No -i input filename given. See -h help or `man phrep`.");
    exit(66 /*EX_NOINPUT*/);
  }
  try {
    $proc->pragma["output"] = 1;  // just in case of eager pre-includes
    $src = $proc->on($opts->input);
  }
  catch (Exception $e) {
    $proc->stderr($e->getCode(), $e->getMessage());
    exit(70 /*EX_SOFTWARE*/);
  }


  // Write to target filename / or stdout
  switch($output_fn = $opts->output) {
    case "/dev/stdout":
    case "-":
    case "":
      print $src;
      exit(0 /*EX_OK*/);
    case "*.php":
    case ".php":
    case "php":
      $output_fn = preg_replace("/(\.[pm])?\.\w{2,4}$/", ".php", $opts->input);
      goto writefile;
    case "input":
    case "-i":
    case "i":
      $output_fn = $opts->input;
      goto writefile;
    case "target":
    case "t":
      $output_fn = $proc->pragma["target"];
    default:
    writefile:
      if (file_exists($output_fn) && !$opts->force) {
        $proc->stderr(2, "Output filename '$output_fn' exists. (Use -f option to force overwriting.)");
        exit(73 /*EX_CANTCREAT*/);
      }
      else {
        file_put_contents($output_fn, $src);
        exit(0 /*EX_OK*/);
      }
  }




  /**
   * Ties getopt($argv) results list to MacroProcessor options,
   * Populates $proc->defines/dirs/pragma fields.
   *
   */
  class ShellFlags {

      static $map = [
          "i" => "input:", "o" => "output:", "f" => "force",
          "I" => "dirs:", "D" => "defines:", "p" => "pragma:", 
          "m" => "mode:", "k" => "keep", "w" => "quiet", "W" => "fail",
          "E" => "cpp", "config" => "config:", "n" => "no:",
          "v" => "verbose", "V" => "version", "h" => "help", 
      ];
      var $input = NULL;
      var $output = "-";
      var $force = false;
      var $proc;


      // Dispatch -flags to handlers
      public function __construct($opts, $proc) {
          $this->proc = $proc;
          foreach ($opts as $key=>$list) {
              // Store literal flag value (mixed str/array)
              in_array($key, self::$map) or $key = trim(self::$map[$key], ":");
              $this->$key = $list;
              // Invoke custom handlers
              foreach ((array)$list as $val) {
                  $this->$key($val, $proc);
              }
          }
      }

      
      // Read config file from HOME/XDG_CONFIG_HOME
      public function config($fn = "phrep/config.php") {
          $proc = $this->proc;
          $opts = $this;
          if (file_exists($fn)) {
              include_once($fn);
          }
          foreach (["XDG_CONFIG_HOME" => $fn, "HOME" => ".config/$fn", "APPDATA" => $fn] as $env => $fn) {
              if (!empty($_SERVER[$env]) and file_exists($fn = "{$_SERVER[$env]}/$fn")) {
                   // config include script which may adapt options
                   return include_once(realpath($fn));
              }
          }
      }


      // Include search dirs, or preread defines.h without output
      function dirs($val) {
          if (file_exists($val)) {
              if (is_dir($val)) {
                  $this->proc->dirs[] = $val;   // add to #include <paths/*>
              }
              else {
                  $this->proc->on($val);   // preinclude.h without output
              }
          }
          else {
              $this->proc->stderr(E_ERROR, "Input file/dir '$val' does not exist.");
          }
      }

    
      // Custom constants
      function defines($val) {
          $this->proc->d_define($val);
      }


      // #Pragma(key=value)
      function pragma($val) {
          $this->proc->d_pragma("($val)");
      }


      // Input filename
      function input($val) {
          $this->proc->pragma["input"] = $val;
          if ($val == "-") {
              $val = "/dev/stdin";
              $this->input = $val;
          }
          else {
              // Also add basedir of input file to #include search $dirs for simplicity
              if ($dir = basename($val)) {
                  $this->proc->dirs[] = $val;
              }
          }
      }


      // Target filename
      function output($val) {
          $this->proc->pragma["target"] = $val;
      }


      // No: s=substitution, or just c=constants, m=macros, x=complex
      function no($val) {
          foreach (str_split($val) as $what) {
              switch ($what) {
                  case "s": $this->proc->pragma["substitution"] = 0; break;
                  case "c": $this->proc->pragma["constants"] = 0; break;
                  case "m": $this->proc->pragma["macros"] = 0; break;
                  case "x": $this->proc->pragma["complex"] = 0; break;
              }
          }
      }


      // Quiet quiet
      function quiet($val) {     $this->proc->pragma["quiet"] = 1;               }

      // Errors
      function fail($val) {      $this->proc->pragma["fail"] = 1;                }

      // Keep lines
      function keep($val) {      $this->proc->pragma["keep_empty"] = 1;          }

      // Interpolation mode (token/regex/erb/phpp)
      function mode($val) {      $this->proc->pragma["interpolate"] = $val;      }

      // verbose
      function verbose($val) {   $this->proc->pragma["verbose"] = $val;          }

      // Version
      function version() {
          preg_match("/version:\s*(\d+\.\S+)/i", file_get_contents(__FILE__), $ver);
          die("phrep $ver[1]\n");
      }


      // Local flags
      function force() { $this->force = 1; }
      function cpp() {  }

      
      // Helpsme
      function help() {
          $h = "\e[1mPHP source file preproccessor\e[0m. Handles #if, #ifdef, #define, #include directives
to substitute constants and macros, or conjoin 'header' files (include.ph).

Usage:
   \e[1;32mphrep\e[0m [-opts] -i source.php [-o target.php]

Options:
   -i source.fn   Source script filename, or pipe input via -i-
   -o filename    Output target filename, or special name -oi, -o input, -o target
   -f             Force overwriting if output file exists.
   -D CONST=VAL   Predefine constant or macro.
   -I dir/        Include directory and search path.
   -I preload.h   Read macros/constants from header file beforehand, without output.
   -E             Dummy CPP compat flag (\"just preprocess\").
   -w             Supress any non-fatal warnings.
   -W             Abort with errors instead of just printing warnings.
   -m token       Switch to `regex` or `phpp` or `erb` mode instead of `tokenizer`.
                  Define custom const/macro delimiters with `-m 'regex,[[,]]'` e.g.
   -k             Keep empty lines in place for #directives.
   -p opt=val     Set other processing (#pragma) options.
   -ns            No substitution of constants or macros.
   -nc            No constant substution, but still macro/complex@() replacement.
   -v             Verbose mode. (Dummy flag.)
   -V             Print version.
   -h             Show this help text.

Source files are scanned for macro directives such as #define, #if, #ifdef, #ifndef,
and #else, #elif, #endif, #undef, #pragma(opt=val), #stderr or #include. After all
include files have been joined, any collected constants and macros will be substituted.

If no -o output filename has been given, the processed script will be printed out
instead. But files can also be updated in-place using `-f -oi`.\n\n";
          exit(
            preg_replace( // colorize #directives, -flags w/ opts and `backticked` stuff
              ["/#\w+/", "/(?<=\W)(-\w+\s)(\S+)/", "/`(.+?)`|(?<=\W)-\w+\.?/"],
              ["\e[2m$0\e[0m", "$1\e[1;32m$2\e[0m", "\e[1;31m$0\e[0m"], $h
            )
          );
      }
  }

} # global namespace

#ifdef PKG_COMBINED
#include <macro.php>
#endif
